// muesli-slang.c -*- C -*-
/* Interface to slang
   Copyright (C) 2008, 2009, 2010 University of Limerick

   This file is part of Muesli.
   
   Muesli is free software: you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation, either version 3 of the License, or (at your
   option) any later version.
   
   Muesli is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
   
   You should have received a copy of the GNU General Public License
   along with Muesli.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _MUESLI_SLANG_
#define _MUESLI_SLANG_

#include "../config.h"
#include "muesli.h"

#include <stdio.h>
#include <limits.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>		// todo: do I need this?

static int ambient_transient = 0;

#ifndef CUSTOMIZED

typedef struct slang_state {
  int fred;
} slang_state;

typedef struct slang_table {
  int count;
} slang_table;

typedef struct slang_symbol {
  char *pname;
} slang_symbol;

#define SLANG_VERSION "unknown"

slang_state *slang_create_custom() { return NULL; }

// Getting args from your scripting language stack:
int slang_count_args(slang_state *S) { return 0; }
int slang_number_arg(slang_state *S, int argi) { return 0; }
char *slang_string_arg(slang_state *S, int argi) { return (char*)""; }
int slang_arg_is_string(slang_state *S, int argi) { return 0; }
int slang_arg_is_number(slang_state *S, int argi) { return 0; }

// Putting results back onto your scripting language stack:
void slang_return_string(slang_state *S, const char *string_back) {}
void slang_return_number(slang_state *S, float number_back) {}
void slang_return_boolean(slang_state *S, int int_back) {}
void slang_return_void(slang_state *S) {};

// Handling table / map / alist structures of your scripting language:
slang_table *slang_create_table(slang_state *S, int size) { return NULL; }
void slang_set_table_string(slang_state *S, slang_table *T, char *key, const char *value) {}
void slang_set_table_number(slang_state *S, slang_table *T, char *key, float value) {}
void slang_set_table_boolean(slang_state *S, slang_table *T, char *key, int value) {}

// Make a name / symbol of the scripting language:
slang_symbol *slang_make_symbol(slang_state *S, const char *n) { return NULL; }

// Non-local exits in the scripting language
void slang_throw(slang_state *S, slang_symbol *Y, float n) {}

// Set a global variable of the scripting language to contain a table:
void slang_set_global_variable_table(slang_state *S, slang_symbol *varname, slang_table *table) {}

// Extend the scripting language with an added built-in function:
void slang_register_builtin_function(slang_state *S, const char *funname, void (*function)(slang_state *FS)) {}
#endif

// A specialized non-local exit; could either throw, or just quit the
// application run:
void slang_error(slang_state *S, char *message, char *arg)
{
  fprintf(stderr, "Got error \"%s\", \"%s\"\n", message, arg);
}

static evaluator_interface *slang_interface = NULL;

static char *option_names_var_name = (char*)"option_names";

static unsigned int scratch_len = 0;
static char* scratch_buf = NULL;

void
slang_set_parameter(slang_state *cs)
{
  if (slang_count_args(cs) < 2) {
    fprintf(stderr, "too few args to slang_set_parameter\n");
    slang_error(cs, (char*)"argcount", (char*)"set_parameter");
  }

  char option_code = muesli_find_option_letter(slang_interface->getopt_options, slang_string_arg(cs, 0));

  if (option_code != -1) {
    eval_result_t result;
    result_type_t result_type;
    if (slang_arg_is_string(cs, 1)) {
      const char *from_custom = slang_string_arg(cs, 1);
      if (strlen(from_custom)+1 > scratch_len) {
	scratch_len = strlen(from_custom)+1;
	if (scratch_buf != NULL) {
	  free(scratch_buf);
	}
	scratch_buf = (char*)malloc(scratch_len);
      }
      strcpy(scratch_buf, from_custom);
      (slang_interface->handle_option)(slang_interface->app_params,
					option_code, scratch_buf, 0.0, 1,
					&result_type, &result, "custom");
    } else if (slang_arg_is_number(cs, 1)) {
      (slang_interface->handle_option)(slang_interface->app_params,
					option_code, NULL, slang_number_arg(cs, 2), 1,
					&result_type, &result, "custom");
    } else {
      (slang_interface->handle_option)(slang_interface->app_params,
					option_code, (char*)"true", 0.0, 1,
					&result_type, &result, "custom");
    }
  }
}

void
slang_set_parameters(slang_state *cs)
{
  if (slang_count_args(cs) < 1)  {
    fprintf(stderr, "too few args to slang_set_parameters\n");
    slang_error(cs, (char*)"argcount", (char*)"set_parameters");
  }

#if 0
  // Fill in: use a table iterator from your language
  slang_table table = slang_table_arg(cs, 0);

  slang_table_iteration_start(cs, table);
  while (slang_table_iteration_next(cs, table) != 0) {
    slang_set_parameter(cs,
			 slang_table_iteration_current_key(cs, table),
			 slang_table_iteration_current_value(cs, table));
  }

#endif
}

void
slang_get_parameter(slang_state *cs)
{
  if (slang_count_args(cs) < 1)  {
    fprintf(stderr, "too few args to slang_get_parameter\n");
    slang_error(cs, (char*)"argcount", (char*)"get_parameter");
  }

  char option_code = muesli_find_option_letter(slang_interface->getopt_options,
					       slang_string_arg(cs, 0));

  if (option_code != -1) {

    result_type_t result_type;
    eval_result_t result;

    (slang_interface->handle_option)(slang_interface->app_params,
				      option_code,	// option
				      NULL, 0.0,	// value
				      0,	// set
				      &result_type, &result,
				      "custom");

    switch (result_type) {
    case string_result:
      slang_return_string(cs, result.as_string);
      break;
    case number_result:
      slang_return_number(cs, result.as_number);
      break;
    case boolean_result:
      slang_return_boolean(cs, result.as_int);
      break;
    default:
      slang_return_boolean(cs, 0);
      break;
    }
  }
}

void
slang_get_parameters(slang_state *cs)
{
  slang_table *table = slang_create_table(cs, 48);

  // todo: fix and re-instate -- I have to get long_options across to it somehow
  struct option *option_names = slang_interface->getopt_options;

  while (option_names->name != 0) {
    
    result_type_t result_type;
    eval_result_t result;

    (slang_interface->handle_option)(slang_interface->app_params,
				      (option_names->val), // opt
				      NULL, 0.0,		     // value
				      0,		     // set
				      &result_type, &result,
				      "custom");

    switch (result_type) {
    case string_result:
      slang_set_table_string(cs, table, (char*)(option_names->name), result.as_string);
      break;
    case number_result:
      slang_set_table_number(cs, table, (char*)(option_names->name), result.as_number);
      break;
    case boolean_result:
      slang_set_table_boolean(cs, table, (char*)(option_names->name), result.as_int);
      break;
    default:
      slang_set_table_boolean(cs, table, (char*)(option_names->name), 0);
      break;
    }
    option_names++;
  }
}

///////////////////////////////////////
// Call arbitrary evaluators by name //
///////////////////////////////////////

static int
muesli_slang_eval_in_language(slang_state *cs)
{
  if (slang_count_args(cs) < 2) {
    fprintf(stderr, "too few args to slang_eval_in_language\n");
    slang_error(cs, (char*)"argcount", (char*)"eval_in_language");
    return 0;
  }

  if ((slang_arg_is_string(cs, 1)) && (slang_arg_is_string(cs, 2))) {
    const char *language_name = slang_string_arg(cs, 1);
    const char *evaluand = slang_string_arg(cs, 2);
    unsigned int evaluand_length = strlen(evaluand);

    fprintf(stderr, "In slang_eval_in_language(\"%s\", \"%s\")\n", language_name, evaluand);

    eval_result_t result;

    result_type_t result_type = muesli_eval_in_language(language_name,
							evaluand,
							evaluand_length,
							&result,
							ambient_transient);

    switch (result_type) {
    case number_result:
      slang_return_number(cs, result.as_number);
      break;
    case string_result:
      slang_return_string(cs, result.as_string);
      break;
    case boolean_result:
      slang_return_boolean(cs, result.as_int);
      break;
    case no_result:
    case error_result:
      slang_return_boolean(cs, 0);
      break;
    }
  } else {
    fprintf(stderr, "wrong type args to slang_eval_in_language\n");
    slang_error(cs, (char*)"argtype", (char*)"eval_in_language");
    return 0;
  }
  return 1;
}

///////////////////////////////
// Custom built-in functions //
///////////////////////////////

// This isn't meant as self-referentially as it looks (pity!).  It's
// really there so that if you use this file as a template for another
// interpreter interface, the language you're adding will get access
// to custom built-in functions just like the existing languages do.

static void
slang_eval_function(slang_state *cs)
{
  if (slang_arg_is_string(cs, 0)) {
    char *string_arg = slang_string_arg(cs, 0);
    eval_result_t result;
    result_type_t result_type = (slang_interface->eval_string)(slang_interface,
								string_arg, strlen(string_arg),
								&result, ambient_transient);
    switch (result_type) {
    case number_result:
      slang_return_number(cs, result.as_number);
      break;
    case string_result:
      slang_return_string(cs, result.as_string);
      break;
    case boolean_result:
      slang_return_boolean(cs, result.as_int);
      break;
    case no_result:
    case error_result:
      slang_return_void(cs);
      break;
    }
  } else {
    fprintf(stderr, "slang_eval_function must be given a string\n");
    slang_error(cs, (char*)"argtype", (char*)"slang_eval");
  }
}

////////////////
// Initialize //
////////////////

void
slang_evaluator_init(evaluator_interface *interface)
{
  // Init custom interface

  // Fill in: initialize your custom application
  slang_state *our_slang_state = slang_create_custom();
  interface->state = our_slang_state;
  slang_interface = interface;

  // Extend system

  // Fill in: register all these language extensions (as applicable):
  slang_register_builtin_function(our_slang_state, (char*)"get_parameter", slang_get_parameter);
  slang_register_builtin_function(our_slang_state, (char*)"set_parameter", slang_set_parameter);
  slang_register_builtin_function(our_slang_state, (char*)"evolution_parameters", slang_get_parameters);
  slang_register_builtin_function(our_slang_state, (char*)"modify_evolution_parameters", slang_set_parameters);
  slang_register_builtin_function(our_slang_state, (char*)"custom", slang_eval_function);

  // Set up option names

  // Fill in: Create a table of option names.  You may well not need to bother.
  slang_table *option_names_table = slang_create_table(our_slang_state, 48);
  struct option *option_names;
  for (option_names = slang_interface->getopt_options;
       option_names->name != 0;
       option_names++) {
    slang_set_table_number(our_slang_state, option_names_table,
			    (char*)(option_names->name), (option_names->val));
  }
  slang_set_global_variable_table(our_slang_state,
				   slang_make_symbol(our_slang_state,
						      option_names_var_name),
				   option_names_table);
}

void
slang_load_file(evaluator_interface *interface,
		const char *filename)
{
  int muesli_flags = interface->flags;
  if (muesli_flags & TRACE_MUESLI_LOAD) {
    fprintf(stderr, "Loading %s\n", filename);
  }
  // Fill in: load the functions file given as (char*)(filename)
  if (muesli_flags & TRACE_MUESLI_LOAD) {
    fprintf(stderr, "Loaded %s\n", filename);
  }
}

result_type_t
slang_eval_string(evaluator_interface *evaluator,
		   const char *slang_c_string,
		   unsigned int string_length,
		   eval_result_t *result_p,
		   int transient)
{
  result_type_t result_type = no_result;

  if (slang_c_string) {
    int old_ambient_transient = ambient_transient;
    ambient_transient = transient;
    // fprintf(stderr, "Custom evaluating string \"%s\"\n", slang_c_string);

    // Fill in: evaluate slang_c_string, decide what kind of result
    // it returns, and fill in the result pointer and the result type
    // accordingly.



    ambient_transient = old_ambient_transient;
  }

  return result_type;
}

void
slang_setup(evaluator_interface *new_slang_interface)
{
  slang_interface = new_slang_interface;

  slang_interface->eval_string = slang_eval_string;
  slang_interface->load_file = slang_load_file;

  slang_interface->version = SLANG_VERSION;
}

#endif

